@@ -1,7 +1,7 @@
 /***********************************************************************/
-/*  Name: ODBCONN.CPP  Version 2.4                                     */
+/*  Name: ODBCONN.CPP  Version 2.3                                     */
 /*                                                                     */
-/*  (C) Copyright to the author Olivier BERTRAND          1998-2021    */
+/*  (C) Copyright to the author Olivier BERTRAND          1998-2017    */
 /*                                                                     */
 /*  This file contains the ODBC connection classes functions.          */
 /***********************************************************************/
@@ -1180,7 +1180,7 @@ int ODBConn::Open(PCSZ ConnectString, POPARM sop, DWORD options)
 //  VerifyConnect();         Deprecated
     GetConnectInfo();
   } catch(DBX *xp) {
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", xp->m_Msg, xp->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", xp->m_Msg, xp->GetErrorMessage(0));
     Close();
 //  Free();
     return -1;
@@ -1289,10 +1289,7 @@ bool ODBConn::DriverConnect(DWORD Options)
 #else   // !_WIN32
   HWND    hWnd = (HWND)1;
 #endif  // !_WIN32
-  PGLOBAL& g = m_G;
-  PDBUSER dup = PlgGetUser(g);
 
-//if (Options & noOdbcDialog || dup->Remote)
     wConnectOption = SQL_DRIVER_NOPROMPT;
 //else if (Options & forceOdbcDialog)
 //  wConnectOption = SQL_DRIVER_PROMPT;
@@ -1422,7 +1419,7 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
   PGLOBAL& g = m_G;
   void    *buffer;
   bool     b;
-  UWORD    n, k;
+  UWORD    n;
   SWORD    len, tp, ncol = 0;
   ODBCCOL *colp;
   RETCODE  rc;
@@ -1489,16 +1486,15 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
     } else {
       do {
         rc = SQLExecDirect(hstmt, (PUCHAR)sql, SQL_NTS);
-      } while (rc == SQL_STILL_EXECUTING);
+        } while (rc == SQL_STILL_EXECUTING);
 
       if (!Check(rc))
         ThrowDBX(rc, "SQLExecDirect", hstmt);
 
       do {
         rc = SQLNumResultCols(hstmt, &ncol);
-      } while (rc == SQL_STILL_EXECUTING);
+        } while (rc == SQL_STILL_EXECUTING);
 
-      k = 0;    // used for column number
     } // endif Srcdef
 
     for (n = 0, colp = tocols; colp; colp = (PODBCCOL)colp->GetNext())
@@ -1510,7 +1506,7 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
       ThrowDBX(MSG(COL_NUM_MISM));
 
     // Now bind the column buffers
-    for (colp = tocols; colp; colp = (PODBCCOL)colp->GetNext())
+    for (n = 1, colp = tocols; colp; colp = (PODBCCOL)colp->GetNext())
       if (!colp->IsSpecial()) {
         buffer = colp->GetBuffer(m_RowsetSize);
         len = colp->GetBuflen();
@@ -1520,23 +1516,19 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
           sprintf(m_G->Message, MSG(INV_COLUMN_TYPE),
                   colp->GetResultType(), SVP(colp->GetName()));
           ThrowDBX(m_G->Message);
-        } // endif tp
-
-        if (m_Tdb->Srcdef)
-          k = colp->GetIndex();
-        else
-          k++;
+          } // endif tp
 
         if (trace(1))
           htrc("Binding col=%u type=%d buf=%p len=%d slen=%p\n",
-                  k, tp, buffer, len, colp->GetStrLen());
+                  n, tp, buffer, len, colp->GetStrLen());
 
-        rc = SQLBindCol(hstmt, k, tp, buffer, len, colp->GetStrLen());
+        rc = SQLBindCol(hstmt, n, tp, buffer, len, colp->GetStrLen());
 
         if (!Check(rc))
           ThrowDBX(rc, "SQLBindCol", hstmt);
 
-      } // endif colp
+        n++;
+        } // endif pcol
 
   } catch(DBX *x) {
     if (trace(1))
@@ -1646,7 +1638,7 @@ int ODBConn::Fetch(int pos)
       for (int i = 0; i < MAX_NUM_OF_MSG && x->m_ErrMsg[i]; i++)
         htrc(x->m_ErrMsg[i]);
 
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", x->m_Msg, x->GetErrorMessage(0));
     irc = -1;
   } // end try/catch
 
@@ -1682,7 +1674,7 @@ int ODBConn::PrepareSQL(char *sql)
         for (int i = 0; i < MAX_NUM_OF_MSG && x->m_ErrMsg[i]; i++)
           htrc(x->m_ErrMsg[i]);
 
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", x->m_Msg, x->GetErrorMessage(0));
     } // end try/catch
 
     } // endif Mode
@@ -1691,7 +1683,7 @@ int ODBConn::PrepareSQL(char *sql)
     b = false;
 
     if (m_hstmt) {
-      RETCODE rc = SQLFreeStmt(m_hstmt, SQL_CLOSE);
+      SQLFreeStmt(m_hstmt, SQL_CLOSE);
 
       hstmt = m_hstmt;
       m_hstmt = NULL;
@@ -1699,7 +1691,7 @@ int ODBConn::PrepareSQL(char *sql)
       if (m_Tdb->GetAmType() != TYPE_AM_XDBC)
         ThrowDBX(MSG(SEQUENCE_ERROR));
 
-      } // endif m_hstmt
+    } // endif m_hstmt
 
     rc = SQLAllocStmt(m_hdbc, &hstmt);
 
@@ -1728,7 +1720,7 @@ int ODBConn::PrepareSQL(char *sql)
       for (int i = 0; i < MAX_NUM_OF_MSG && x->m_ErrMsg[i]; i++)
         htrc(x->m_ErrMsg[i]);
 
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", x->m_Msg, x->GetErrorMessage(0));
 
     if (b)
       SQLCancel(hstmt);
@@ -1928,7 +1920,7 @@ bool ODBConn::ExecSQLcommand(char *sql)
 			for (int i = 0; i < MAX_NUM_OF_MSG && x->m_ErrMsg[i]; i++)
 				htrc(x->m_ErrMsg[i]);
 
-    snprintf(g->Message, sizeof(g->Message), "Remote %s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "Remote %s: %s", x->m_Msg, x->GetErrorMessage(0));
 
     if (b)
       SQLCancel(hstmt);
@@ -1938,17 +1930,17 @@ bool ODBConn::ExecSQLcommand(char *sql)
   } // end try/catch
 
   if (!Check(rc = SQLFreeStmt(hstmt, SQL_CLOSE)))
-    snprintf(g->Message, sizeof(g->Message), "SQLFreeStmt: rc=%d", rc);
+    sprintf(g->Message, "SQLFreeStmt: rc=%d", rc);
 
   if (m_Transact) {
     // Terminate the transaction
     if (!Check(rc = SQLEndTran(SQL_HANDLE_DBC, m_hdbc, 
                        (rcd) ? SQL_ROLLBACK : SQL_COMMIT)))
-      snprintf(g->Message, sizeof(g->Message), "SQLEndTran: rc=%d", rc);
+      sprintf(g->Message, "SQLEndTran: rc=%d", rc);
 
     if (!Check(rc = SQLSetConnectAttr(m_hdbc, SQL_ATTR_AUTOCOMMIT,
                (SQLPOINTER)SQL_AUTOCOMMIT_ON, SQL_IS_UINTEGER)))
-      snprintf(g->Message, sizeof(g->Message), "SQLSetConnectAttr: rc=%d", rc);
+      sprintf(g->Message, "SQLSetConnectAttr: rc=%d", rc);
 
     m_Transact = false;
     } // endif m_Transact
@@ -2013,7 +2005,7 @@ PQRYRES ODBConn::GetMetaData(PGLOBAL g, PCSZ dsn, PCSZ src)
       } // endfor i
 
   } catch(DBX *x) {
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", x->m_Msg, x->GetErrorMessage(0));
     goto err;
   } // end try/catch
 
@@ -2064,7 +2056,7 @@ PQRYRES ODBConn::GetMetaData(PGLOBAL g, PCSZ dsn, PCSZ src)
       } // endfor i
 
   } catch(DBX *x) {
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", x->m_Msg, x->GetErrorMessage(0));
     qrp = NULL;
   } // end try/catch
 
@@ -2373,7 +2365,7 @@ int ODBConn::GetCatInfo(CATPARM *cap)
     // Now bind the column buffers
     for (n = 0, crp = qrp->Colresp; crp; crp = crp->Next) {
       if ((tp = GetSQLCType(crp->Type)) == SQL_TYPE_NULL) {
-        snprintf(g->Message, sizeof(g->Message), MSG(INV_COLUMN_TYPE), crp->Type, crp->Name);
+        sprintf(g->Message, MSG(INV_COLUMN_TYPE), crp->Type, crp->Name);
         ThrowDBX(g->Message);
         } // endif tp
 
@@ -2459,9 +2451,9 @@ int ODBConn::GetCatInfo(CATPARM *cap)
 
     if (rc == SQL_NO_DATA_FOUND) {
       if (cap->Pat)
-        snprintf(g->Message, sizeof(g->Message), MSG(NO_TABCOL_DATA), cap->Tab, cap->Pat);
+        sprintf(g->Message, MSG(NO_TABCOL_DATA), cap->Tab, cap->Pat);
       else
-        snprintf(g->Message, sizeof(g->Message), MSG(NO_TAB_DATA), cap->Tab);
+        sprintf(g->Message, MSG(NO_TAB_DATA), cap->Tab);
 
       ThrowDBX(g->Message);
     } else if (rc == SQL_SUCCESS) {
@@ -2486,7 +2478,7 @@ int ODBConn::GetCatInfo(CATPARM *cap)
       for (int i = 0; i < MAX_NUM_OF_MSG && x->m_ErrMsg[i]; i++)
         htrc(x->m_ErrMsg[i]);
 
-    snprintf(g->Message, sizeof(g->Message), "%s: %s", x->m_Msg, x->GetErrorMessage(0));
+    sprintf(g->Message, "%s: %s", x->m_Msg, x->GetErrorMessage(0));
     irc = -1;
   } // end try/catch
 
@@ -2548,7 +2540,7 @@ PQRYRES ODBConn::AllocateResult(PGLOBAL g)
 
       if (!(crp->Kdata = AllocValBlock(g, NULL, crp->Type, m_Rows,
                                      crp->Clen, 0, FALSE, TRUE, uns))) {
-        snprintf(g->Message, sizeof(g->Message), MSG(INV_RESULT_TYPE),
+        sprintf(g->Message, MSG(INV_RESULT_TYPE),
                           GetFormatType(crp->Type));
         return NULL;
         } // endif Kdata
