@@ -2,7 +2,7 @@
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2009, Google Inc.
-Copyright (c) 2014, 2020, MariaDB Corporation.
+Copyright (c) 2014, 2021, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -50,6 +50,7 @@ Created 12/9/1995 Heikki Tuuri
 #include "trx0trx.h"
 #include "trx0roll.h"
 #include "srv0mon.h"
+#include "sync0sync.h"
 #include "buf0dump.h"
 #include "log0sync.h"
 
@@ -63,14 +64,6 @@ to the InnoDB redo log. */
 /** Redo log system */
 log_t	log_sys;
 
-/* These control how often we print warnings if the last checkpoint is too
-old */
-static bool	log_has_printed_chkp_warning = false;
-static time_t	log_last_warning_time;
-
-static bool	log_has_printed_chkp_margine_warning = false;
-static time_t	log_last_margine_warning_time;
-
 /* A margin for free space in the log buffer before a log entry is catenated */
 #define LOG_BUF_WRITE_MARGIN	(4 * OS_FILE_LOG_BLOCK_SIZE)
 
@@ -79,31 +72,6 @@ static time_t	log_last_margine_warning_time;
 #define LOG_BUF_FLUSH_MARGIN	(LOG_BUF_WRITE_MARGIN		\
 				 + (4U << srv_page_size_shift))
 
-/* This parameter controls asynchronous making of a new checkpoint; the value
-should be bigger than LOG_POOL_PREFLUSH_RATIO_SYNC */
-
-#define LOG_POOL_CHECKPOINT_RATIO_ASYNC	32
-
-/* This parameter controls synchronous preflushing of modified buffer pages */
-#define LOG_POOL_PREFLUSH_RATIO_SYNC	16
-
-/* The same ratio for asynchronous preflushing; this value should be less than
-the previous */
-#define LOG_POOL_PREFLUSH_RATIO_ASYNC	8
-
-/** Return the oldest modified LSN in buf_pool.flush_list,
-or the latest LSN if all pages are clean.
-@return LSN of oldest modification */
-static lsn_t log_buf_pool_get_oldest_modification()
-{
-  ut_ad(log_mutex_own());
-  log_flush_order_mutex_enter();
-  lsn_t lsn= buf_pool.get_oldest_modification();
-  log_flush_order_mutex_exit();
-
-  return lsn ? lsn : log_sys.get_lsn();
-}
-
 /** Extends the log buffer.
 @param[in]	len	requested minimum size in bytes */
 void log_buffer_extend(ulong len)
@@ -116,11 +84,11 @@ void log_buffer_extend(ulong len)
 		(ut_malloc_dontdump(new_buf_size, PSI_INSTRUMENT_ME));
 	TRASH_ALLOC(new_flush_buf, new_buf_size);
 
-	log_mutex_enter();
+	mysql_mutex_lock(&log_sys.mutex);
 
 	if (len <= srv_log_buffer_size) {
 		/* Already extended enough by the others */
-		log_mutex_exit();
+		mysql_mutex_unlock(&log_sys.mutex);
 		ut_free_dodump(new_buf, new_buf_size);
 		ut_free_dodump(new_flush_buf, new_buf_size);
 		return;
@@ -142,7 +110,7 @@ void log_buffer_extend(ulong len)
 	log_sys.max_buf_free = new_buf_size / LOG_BUF_FLUSH_RATIO
 		- LOG_BUF_FLUSH_MARGIN;
 
-	log_mutex_exit();
+	mysql_mutex_unlock(&log_sys.mutex);
 
 	ut_free_dodump(old_buf, old_buf_size);
 	ut_free_dodump(old_flush_buf, old_buf_size);
@@ -151,276 +119,6 @@ void log_buffer_extend(ulong len)
 		<< new_buf_size << ".";
 }
 
-/** Calculate actual length in redo buffer and file including
-block header and trailer.
-@param[in]	len	length to write
-@return actual length to write including header and trailer. */
-static inline
-ulint
-log_calculate_actual_len(
-	ulint len)
-{
-	ut_ad(log_mutex_own());
-
-	const ulint	framing_size = log_sys.framing_size();
-	/* actual length stored per block */
-	const ulint	len_per_blk = OS_FILE_LOG_BLOCK_SIZE - framing_size;
-
-	/* actual data length in last block already written */
-	ulint	extra_len = (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE);
-
-	ut_ad(extra_len >= LOG_BLOCK_HDR_SIZE);
-	extra_len -= LOG_BLOCK_HDR_SIZE;
-
-	/* total extra length for block header and trailer */
-	extra_len = ((len + extra_len) / len_per_blk) * framing_size;
-
-	return(len + extra_len);
-}
-
-/** Check margin not to overwrite transaction log from the last checkpoint.
-If would estimate the log write to exceed the log_capacity,
-waits for the checkpoint is done enough.
-@param[in]	len	length of the data to be written */
-
-void
-log_margin_checkpoint_age(
-	ulint	len)
-{
-	ulint	margin = log_calculate_actual_len(len);
-
-	ut_ad(log_mutex_own());
-
-	if (margin > log_sys.log_capacity) {
-		/* return with warning output to avoid deadlock */
-		if (!log_has_printed_chkp_margine_warning
-		    || difftime(time(NULL),
-				log_last_margine_warning_time) > 15) {
-			log_has_printed_chkp_margine_warning = true;
-			log_last_margine_warning_time = time(NULL);
-
-			ib::error() << "The transaction log file is too"
-				" small for the single transaction log (size="
-				<< len << "). So, the last checkpoint age"
-				" might exceed the log capacity "
-				<< log_sys.log_capacity << ".";
-		}
-
-		return;
-	}
-
-	/* Our margin check should ensure that we never reach this condition.
-	Try to do checkpoint once. We cannot keep waiting here as it might
-	result in hang in case the current mtr has latch on oldest lsn */
-	const lsn_t lsn = log_sys.get_lsn();
-
-	if (lsn - log_sys.last_checkpoint_lsn + margin
-	    > log_sys.log_capacity) {
-		/* The log write of 'len' might overwrite the transaction log
-		after the last checkpoint. Makes checkpoint. */
-
-		const bool flushed_enough = lsn
-			- log_buf_pool_get_oldest_modification() + margin
-			<= log_sys.log_capacity;
-
-		log_sys.set_check_flush_or_checkpoint();
-		log_mutex_exit();
-
-		DEBUG_SYNC_C("margin_checkpoint_age_rescue");
-
-		if (!flushed_enough) {
-			os_thread_sleep(100000);
-		}
-		log_checkpoint();
-
-		log_mutex_enter();
-	}
-
-	return;
-}
-
-/** Open the log for log_write_low. The log must be closed with log_close.
-@param[in]	len	length of the data to be written
-@return start lsn of the log record */
-lsn_t
-log_reserve_and_open(
-	ulint	len)
-{
-	ulint	len_upper_limit;
-#ifdef UNIV_DEBUG
-	ulint	count			= 0;
-#endif /* UNIV_DEBUG */
-
-loop:
-	ut_ad(log_mutex_own());
-
-	/* Calculate an upper limit for the space the string may take in the
-	log buffer */
-
-	len_upper_limit = LOG_BUF_WRITE_MARGIN + srv_log_write_ahead_size
-			  + (5 * len) / 4;
-
-	if (log_sys.buf_free + len_upper_limit > srv_log_buffer_size) {
-		log_mutex_exit();
-
-		DEBUG_SYNC_C("log_buf_size_exceeded");
-
-		/* Not enough free space, do a write of the log buffer */
-		log_sys.initiate_write(false);
-
-		srv_stats.log_waits.inc();
-
-		ut_ad(++count < 50);
-
-		log_mutex_enter();
-		goto loop;
-	}
-
-	return(log_sys.get_lsn());
-}
-
-/************************************************************//**
-Writes to the log the string given. It is assumed that the caller holds the
-log mutex. */
-void
-log_write_low(
-/*==========*/
-	const byte*	str,		/*!< in: string */
-	ulint		str_len)	/*!< in: string length */
-{
-	ulint	len;
-
-	ut_ad(log_mutex_own());
-	const ulint trailer_offset = log_sys.trailer_offset();
-part_loop:
-	/* Calculate a part length */
-
-	ulint data_len = (log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE) + str_len;
-
-	if (data_len <= trailer_offset) {
-
-		/* The string fits within the current log block */
-
-		len = str_len;
-	} else {
-		data_len = trailer_offset;
-
-		len = trailer_offset
-			- log_sys.buf_free % OS_FILE_LOG_BLOCK_SIZE;
-	}
-
-	memcpy(log_sys.buf + log_sys.buf_free, str, len);
-
-	str_len -= len;
-	str = str + len;
-
-	byte* log_block = static_cast<byte*>(
-		ut_align_down(log_sys.buf + log_sys.buf_free,
-			      OS_FILE_LOG_BLOCK_SIZE));
-
-	log_block_set_data_len(log_block, data_len);
-	lsn_t lsn = log_sys.get_lsn();
-
-	if (data_len == trailer_offset) {
-		/* This block became full */
-		log_block_set_data_len(log_block, OS_FILE_LOG_BLOCK_SIZE);
-		log_block_set_checkpoint_no(log_block,
-					    log_sys.next_checkpoint_no);
-		len += log_sys.framing_size();
-
-		lsn += len;
-
-		/* Initialize the next block header */
-		log_block_init(log_block + OS_FILE_LOG_BLOCK_SIZE, lsn);
-	} else {
-		lsn += len;
-	}
-
-	log_sys.set_lsn(lsn);
-	log_sys.buf_free += len;
-
-	ut_ad(log_sys.buf_free <= size_t{srv_log_buffer_size});
-
-	if (str_len > 0) {
-		goto part_loop;
-	}
-
-	srv_stats.log_write_requests.inc();
-}
-
-/************************************************************//**
-Closes the log.
-@return lsn */
-lsn_t
-log_close(void)
-/*===========*/
-{
-	byte*		log_block;
-	ulint		first_rec_group;
-	lsn_t		oldest_lsn;
-	lsn_t		lsn;
-	lsn_t		checkpoint_age;
-
-	ut_ad(log_mutex_own());
-
-	lsn = log_sys.get_lsn();
-
-	log_block = static_cast<byte*>(
-		ut_align_down(log_sys.buf + log_sys.buf_free,
-			      OS_FILE_LOG_BLOCK_SIZE));
-
-	first_rec_group = log_block_get_first_rec_group(log_block);
-
-	if (first_rec_group == 0) {
-		/* We initialized a new log block which was not written
-		full by the current mtr: the next mtr log record group
-		will start within this block at the offset data_len */
-
-		log_block_set_first_rec_group(
-			log_block, log_block_get_data_len(log_block));
-	}
-
-	if (log_sys.buf_free > log_sys.max_buf_free) {
-		log_sys.set_check_flush_or_checkpoint();
-	}
-
-	checkpoint_age = lsn - log_sys.last_checkpoint_lsn;
-
-	if (checkpoint_age >= log_sys.log_capacity) {
-		DBUG_EXECUTE_IF(
-			"print_all_chkp_warnings",
-			log_has_printed_chkp_warning = false;);
-
-		if (!log_has_printed_chkp_warning
-		    || difftime(time(NULL), log_last_warning_time) > 15) {
-
-			log_has_printed_chkp_warning = true;
-			log_last_warning_time = time(NULL);
-
-			ib::error() << "The age of the last checkpoint is "
-				    << checkpoint_age
-				    << ", which exceeds the log capacity "
-				    << log_sys.log_capacity << ".";
-		}
-	}
-
-	if (checkpoint_age <= log_sys.max_modified_age_sync ||
-	    log_sys.check_flush_or_checkpoint()) {
-		goto function_exit;
-	}
-
-	oldest_lsn = log_buf_pool_get_oldest_modification();
-
-	if (!oldest_lsn
-	    || lsn - oldest_lsn > log_sys.max_modified_age_sync
-	    || checkpoint_age > log_sys.max_checkpoint_age_async) {
-		log_sys.set_check_flush_or_checkpoint();
-	}
-function_exit:
-
-	return(lsn);
-}
-
 /** Calculate the recommended highest values for lsn - last_checkpoint_lsn
 and lsn - buf_pool.get_oldest_modification().
 @param[in]	file_size	requested innodb_log_file_size
@@ -461,20 +159,14 @@ log_set_capacity(ulonglong file_size)
 	margin = smallest_capacity - free;
 	margin = margin - margin / 10;	/* Add still some extra safety */
 
-	log_mutex_enter();
+	mysql_mutex_lock(&log_sys.mutex);
 
 	log_sys.log_capacity = smallest_capacity;
 
-	log_sys.max_modified_age_async = margin
-		- margin / LOG_POOL_PREFLUSH_RATIO_ASYNC;
-	log_sys.max_modified_age_sync = margin
-		- margin / LOG_POOL_PREFLUSH_RATIO_SYNC;
-
-	log_sys.max_checkpoint_age_async = margin - margin
-		/ LOG_POOL_CHECKPOINT_RATIO_ASYNC;
+	log_sys.max_modified_age_async = margin - margin / 8;
 	log_sys.max_checkpoint_age = margin;
 
-	log_mutex_exit();
+	mysql_mutex_unlock(&log_sys.mutex);
 
 	return(true);
 }
@@ -486,14 +178,14 @@ void log_t::create()
   ut_ad(!is_initialised());
   m_initialised= true;
 
-  mutex_create(LATCH_ID_LOG_SYS, &mutex);
-  mutex_create(LATCH_ID_LOG_FLUSH_ORDER, &log_flush_order_mutex);
+  mysql_mutex_init(log_sys_mutex_key, &mutex, nullptr);
+  mysql_mutex_init(log_flush_order_mutex_key, &flush_order_mutex, nullptr);
 
   /* Start the lsn from one log block from zero: this way every
   log record has a non-zero start lsn, a fact which we will use */
 
   set_lsn(LOG_START_LSN + LOG_BLOCK_HDR_SIZE);
-  set_flushed_lsn(0);
+  set_flushed_lsn(LOG_START_LSN + LOG_BLOCK_HDR_SIZE);
 
   ut_ad(srv_log_buffer_size >= 16 * OS_FILE_LOG_BLOCK_SIZE);
   ut_ad(srv_log_buffer_size >= 4U << srv_page_size_shift);
@@ -518,8 +210,6 @@ void log_t::create()
   n_log_ios_old= 0;
   log_capacity= 0;
   max_modified_age_async= 0;
-  max_modified_age_sync= 0;
-  max_checkpoint_age_async= 0;
   max_checkpoint_age= 0;
   next_checkpoint_no= 0;
   next_checkpoint_lsn= 0;
@@ -529,6 +219,8 @@ void log_t::create()
   log_block_set_first_rec_group(buf, LOG_BLOCK_HDR_SIZE);
 
   buf_free= LOG_BLOCK_HDR_SIZE;
+  checkpoint_buf= static_cast<byte*>
+    (aligned_malloc(OS_FILE_LOG_BLOCK_SIZE, OS_FILE_LOG_BLOCK_SIZE));
 }
 
 mapped_file_t::~mapped_file_t() noexcept
@@ -771,8 +463,8 @@ void log_t::file::write_header_durable(lsn_t lsn)
   ut_ad(log_sys.log.format == log_t::FORMAT_10_5 ||
         log_sys.log.format == log_t::FORMAT_ENC_10_5);
 
-  // man 2 open suggests this buffer to be aligned by 512 for O_DIRECT
-  MY_ALIGNED(OS_FILE_LOG_BLOCK_SIZE) byte buf[OS_FILE_LOG_BLOCK_SIZE] = {0};
+  byte *buf= log_sys.checkpoint_buf;
+  memset_aligned<OS_FILE_LOG_BLOCK_SIZE>(buf, 0, OS_FILE_LOG_BLOCK_SIZE);
 
   mach_write_to_4(buf + LOG_HEADER_FORMAT, log_sys.log.format);
   mach_write_to_4(buf + LOG_HEADER_SUBFORMAT, log_sys.log.subformat);
@@ -785,7 +477,7 @@ void log_t::file::write_header_durable(lsn_t lsn)
 
   DBUG_PRINT("ib_log", ("write " LSN_PF, lsn));
 
-  log_sys.log.write(0, buf);
+  log_sys.log.write(0, {buf, OS_FILE_LOG_BLOCK_SIZE});
   if (!log_sys.log.writes_are_durable())
     log_sys.log.flush();
 }
@@ -823,11 +515,12 @@ void log_t::file::flush()
 
 void log_t::file::close_file()
 {
-  if (!fd.is_opened())
-    return;
-
-  if (const dberr_t err= fd.close())
-    ib::fatal() << "close(" << fd.get_path() << ") returned " << err;
+  if (fd.is_opened())
+  {
+    if (const dberr_t err= fd.close())
+      ib::fatal() << "close(" << fd.get_path() << ") returned " << err;
+  }
+  fd.free();                                    // Free path
 }
 
 /** Initialize the redo log. */
@@ -932,10 +625,11 @@ loop:
 }
 
 /** Flush the recently written changes to the log file.
-and invoke log_mutex_enter(). */
+and invoke mysql_mutex_lock(&log_sys.mutex). */
 static void log_write_flush_to_disk_low(lsn_t lsn)
 {
-  log_sys.log.flush();
+  if (!log_sys.log.writes_are_durable())
+    log_sys.log.flush();
   ut_a(lsn >= log_sys.get_flushed_lsn());
   log_sys.set_flushed_lsn(lsn);
 }
@@ -947,7 +641,7 @@ static inline
 void
 log_buffer_switch()
 {
-	ut_ad(log_mutex_own());
+	mysql_mutex_assert_owner(&log_sys.mutex);
 	ut_ad(log_write_lock_own());
 
 	size_t		area_end = ut_calc_align<size_t>(
@@ -965,24 +659,28 @@ log_buffer_switch()
 	log_sys.buf_next_to_write = log_sys.buf_free;
 }
 
+/** Invoke commit_checkpoint_notify_ha() to notify that outstanding
+log writes have been completed. */
+void log_flush_notify(lsn_t flush_lsn);
+
 /**
 Writes log buffer to disk
 which is the "write" part of log_write_up_to().
 
 This function does not flush anything.
 
-Note : the caller must have log_mutex locked, and this
+Note : the caller must have log_sys.mutex locked, and this
 mutex is released in the function.
 
 */
 static void log_write(bool rotate_key)
 {
-	ut_ad(log_mutex_own());
+	mysql_mutex_assert_owner(&log_sys.mutex);
 	ut_ad(!recv_no_log_write);
 	lsn_t write_lsn;
 	if (log_sys.buf_free == log_sys.buf_next_to_write) {
 		/* Nothing to write */
-		log_mutex_exit();
+		mysql_mutex_unlock(&log_sys.mutex);
 		return;
 	}
 
@@ -1019,7 +717,7 @@ static void log_write(bool rotate_key)
 
 	log_sys.log.set_fields(log_sys.write_lsn);
 
-	log_mutex_exit();
+	mysql_mutex_unlock(&log_sys.mutex);
 	/* Erase the end of the last log block. */
 	memset(write_buf + end_offset, 0,
 	       ~end_offset & (OS_FILE_LOG_BLOCK_SIZE - 1));
@@ -1067,8 +765,10 @@ static void log_write(bool rotate_key)
 		start_offset - area_start);
 	srv_stats.log_padded.add(pad_size);
 	log_sys.write_lsn = write_lsn;
-	if (log_sys.log.writes_are_durable())
+	if (log_sys.log.writes_are_durable()) {
 		log_sys.set_flushed_lsn(write_lsn);
+		log_flush_notify(write_lsn);
+	}
 	return;
 }
 
@@ -1094,6 +794,7 @@ void log_write_up_to(lsn_t lsn, bool flush_to_disk, bool rotate_key)
 {
   ut_ad(!srv_read_only_mode);
   ut_ad(!rotate_key || flush_to_disk);
+  ut_ad(lsn != LSN_MAX);
 
   if (recv_no_ibuf_operations)
   {
@@ -1110,7 +811,7 @@ void log_write_up_to(lsn_t lsn, bool flush_to_disk, bool rotate_key)
 
   if (write_lock.acquire(lsn) == group_commit_lock::ACQUIRED)
   {
-    log_mutex_enter();
+    mysql_mutex_lock(&log_sys.mutex);
     lsn_t write_lsn= log_sys.get_lsn();
     write_lock.set_pending(write_lsn);
 
@@ -1128,117 +829,47 @@ void log_write_up_to(lsn_t lsn, bool flush_to_disk, bool rotate_key)
   /* Flush the highest written lsn.*/
   auto flush_lsn = write_lock.value();
   flush_lock.set_pending(flush_lsn);
-
-  if (!log_sys.log.writes_are_durable())
-  {
-    log_write_flush_to_disk_low(flush_lsn);
-  }
-
+  log_write_flush_to_disk_low(flush_lsn);
   flush_lock.release(flush_lsn);
 
-  innobase_mysql_log_notify(flush_lsn);
+  log_flush_notify(flush_lsn);
 }
 
 /** write to the log file up to the last log entry.
 @param[in]	sync	whether we want the written log
 also to be flushed to disk. */
-void
-log_buffer_flush_to_disk(
-	bool sync)
+void log_buffer_flush_to_disk(bool sync)
 {
-	ut_ad(!srv_read_only_mode);
-	log_write_up_to(log_get_lsn(), sync);
+  ut_ad(!srv_read_only_mode);
+  log_write_up_to(log_sys.get_lsn(std::memory_order_acquire), sync);
 }
 
 /********************************************************************
 
 Tries to establish a big enough margin of free space in the log buffer, such
 that a new log entry can be catenated without an immediate need for a flush. */
-static
-void
-log_flush_margin(void)
-/*==================*/
+ATTRIBUTE_COLD static void log_flush_margin()
 {
 	lsn_t	lsn	= 0;
 
-	log_mutex_enter();
+	mysql_mutex_lock(&log_sys.mutex);
 
 	if (log_sys.buf_free > log_sys.max_buf_free) {
 		/* We can write during flush */
 		lsn = log_sys.get_lsn();
 	}
 
-	log_mutex_exit();
+	mysql_mutex_unlock(&log_sys.mutex);
 
 	if (lsn) {
 		log_write_up_to(lsn, false);
 	}
 }
 
-/** Advances the smallest lsn for which there are unflushed dirty blocks in the
-buffer pool.
-NOTE: this function may only be called if the calling thread owns no
-synchronization objects!
-@param[in]	new_oldest	try to advance oldest_modified_lsn at least to
-this lsn
-@return false if there was a flush batch of the same type running,
-which means that we could not start this flush batch */
-static bool log_preflush_pool_modified_pages(lsn_t new_oldest)
-{
-	bool	success;
-
-	if (recv_recovery_is_on()) {
-		/* If the recovery is running, we must first apply all
-		log records to their respective file pages to get the
-		right modify lsn values to these pages: otherwise, there
-		might be pages on disk which are not yet recovered to the
-		current lsn, and even after calling this function, we could
-		not know how up-to-date the disk version of the database is,
-		and we could not make a new checkpoint on the basis of the
-		info on the buffer pool only. */
-		recv_sys.apply(true);
-	}
-
-	if (new_oldest == LSN_MAX
-	    || !buf_page_cleaner_is_active
-	    || srv_is_being_started) {
-
-		ulint	n_pages;
-
-		success = buf_flush_lists(ULINT_MAX, new_oldest, &n_pages);
-
-		buf_flush_wait_batch_end(false);
-
-		if (!success) {
-			MONITOR_INC(MONITOR_FLUSH_SYNC_WAITS);
-		}
-
-		MONITOR_INC_VALUE_CUMULATIVE(
-			MONITOR_FLUSH_SYNC_TOTAL_PAGE,
-			MONITOR_FLUSH_SYNC_COUNT,
-			MONITOR_FLUSH_SYNC_PAGES,
-			n_pages);
-	} else {
-		/* better to wait for flushed by page cleaner */
-
-		if (srv_flush_sync) {
-			/* wake page cleaner for IO burst */
-			buf_flush_request_force(new_oldest);
-		}
-
-		buf_flush_wait_flushed(new_oldest);
-
-		success = true;
-	}
-
-	return(success);
-}
-
-/** Write checkpoint info to the log header and invoke log_mutex_exit().
+/** Write checkpoint info to the log header and release log_sys.mutex.
 @param[in]	end_lsn	start LSN of the FILE_CHECKPOINT mini-transaction */
-void log_write_checkpoint_info(lsn_t end_lsn)
+ATTRIBUTE_COLD void log_write_checkpoint_info(lsn_t end_lsn)
 {
-	ut_ad(log_mutex_own());
 	ut_ad(!srv_read_only_mode);
 	ut_ad(end_lsn == 0 || end_lsn >= log_sys.next_checkpoint_lsn);
 	ut_ad(end_lsn <= log_sys.get_lsn());
@@ -1251,7 +882,7 @@ void log_write_checkpoint_info(lsn_t end_lsn)
 			      log_sys.next_checkpoint_lsn));
 
 	byte* buf = log_sys.checkpoint_buf;
-	memset(buf, 0, OS_FILE_LOG_BLOCK_SIZE);
+	memset_aligned<OS_FILE_LOG_BLOCK_SIZE>(buf, 0, OS_FILE_LOG_BLOCK_SIZE);
 
 	mach_write_to_8(buf + LOG_CHECKPOINT_NO, log_sys.next_checkpoint_no);
 	mach_write_to_8(buf + LOG_CHECKPOINT_LSN, log_sys.next_checkpoint_lsn);
@@ -1274,7 +905,7 @@ void log_write_checkpoint_info(lsn_t end_lsn)
 
 	++log_sys.n_pending_checkpoint_writes;
 
-	log_mutex_exit();
+	mysql_mutex_unlock(&log_sys.mutex);
 
 	/* Note: We alternate the physical place of the checkpoint info.
 	See the (next_checkpoint_no & 1) below. */
@@ -1285,7 +916,7 @@ void log_write_checkpoint_info(lsn_t end_lsn)
 
 	log_sys.log.flush();
 
-	log_mutex_enter();
+	mysql_mutex_lock(&log_sys.mutex);
 
 	--log_sys.n_pending_checkpoint_writes;
 	ut_ad(log_sys.n_pending_checkpoint_writes == 0);
@@ -1296,204 +927,50 @@ void log_write_checkpoint_info(lsn_t end_lsn)
 
 	DBUG_PRINT("ib_log", ("checkpoint ended at " LSN_PF
 			      ", flushed to " LSN_PF,
-			      log_sys.last_checkpoint_lsn,
+			      lsn_t{log_sys.last_checkpoint_lsn},
 			      log_sys.get_flushed_lsn()));
 
 	MONITOR_INC(MONITOR_NUM_CHECKPOINT);
 
 	DBUG_EXECUTE_IF("crash_after_checkpoint", DBUG_SUICIDE(););
 
-	log_mutex_exit();
-}
-
-/** Make a checkpoint. Note that this function does not flush dirty
-blocks from the buffer pool: it only checks what is lsn of the oldest
-modification in the pool, and writes information about the lsn in
-log file. Use log_make_checkpoint() to flush also the pool.
-@return true if success, false if a checkpoint write was already running */
-bool log_checkpoint()
-{
-	lsn_t	oldest_lsn;
-
-	ut_ad(!srv_read_only_mode);
-
-	DBUG_EXECUTE_IF("no_checkpoint",
-			/* We sleep for a long enough time, forcing
-			the checkpoint doesn't happen any more. */
-			os_thread_sleep(360000000););
-
-	if (recv_recovery_is_on()) {
-		recv_sys.apply(true);
-	}
-
-	switch (srv_file_flush_method) {
-	case SRV_NOSYNC:
-		break;
-	case SRV_O_DSYNC:
-	case SRV_FSYNC:
-	case SRV_LITTLESYNC:
-	case SRV_O_DIRECT:
-	case SRV_O_DIRECT_NO_FSYNC:
-#ifdef _WIN32
-	case SRV_ALL_O_DIRECT_FSYNC:
-#endif
-		fil_flush_file_spaces();
-	}
-
-	log_mutex_enter();
-
-	ut_ad(!recv_no_log_write);
-	oldest_lsn = log_buf_pool_get_oldest_modification();
-
-	/* Because log also contains headers and dummy log records,
-	log_buf_pool_get_oldest_modification() will return log_sys.lsn
-	if the buffer pool contains no dirty buffers.
-	We must make sure that the log is flushed up to that lsn.
-	If there are dirty buffers in the buffer pool, then our
-	write-ahead-logging algorithm ensures that the log has been
-	flushed up to oldest_lsn. */
-
-	ut_ad(oldest_lsn >= log_sys.last_checkpoint_lsn);
-	if (oldest_lsn
-	    > log_sys.last_checkpoint_lsn + SIZE_OF_FILE_CHECKPOINT) {
-		/* Some log has been written since the previous checkpoint. */
-	} else if (srv_shutdown_state > SRV_SHUTDOWN_INITIATED) {
-		/* MariaDB startup expects the redo log file to be
-		logically empty (not even containing a MLOG_CHECKPOINT record)
-		after a clean shutdown. Perform an extra checkpoint at
-		shutdown. */
-	} else {
-		/* Do nothing, because nothing was logged (other than
-		a FILE_CHECKPOINT marker) since the previous checkpoint. */
-		log_mutex_exit();
-		return(true);
-	}
-	/* Repeat the FILE_MODIFY records after the checkpoint, in
-	case some log records between the checkpoint and log_sys.lsn
-	need them. Finally, write a FILE_CHECKPOINT marker. Redo log
-	apply expects to see a FILE_CHECKPOINT after the checkpoint,
-	except on clean shutdown, where the log will be empty after
-	the checkpoint.
-	It is important that we write out the redo log before any
-	further dirty pages are flushed to the tablespace files.  At
-	this point, because log_mutex_own(), mtr_commit() in other
-	threads will be blocked, and no pages can be added to the
-	flush lists. */
-	lsn_t		flush_lsn	= oldest_lsn;
-	const lsn_t	end_lsn		= log_sys.get_lsn();
-	const bool	do_write
-		= srv_shutdown_state <= SRV_SHUTDOWN_INITIATED
-		|| flush_lsn != end_lsn;
-
-	if (fil_names_clear(flush_lsn, do_write)) {
-		flush_lsn = log_sys.get_lsn();
-		ut_ad(flush_lsn >= end_lsn + SIZE_OF_FILE_CHECKPOINT);
-	}
-
-	log_mutex_exit();
-
-	log_write_up_to(flush_lsn, true, true);
-
-	log_mutex_enter();
-
-	ut_ad(log_sys.get_flushed_lsn() >= flush_lsn);
-	ut_ad(flush_lsn >= oldest_lsn);
-
-	if (log_sys.last_checkpoint_lsn >= oldest_lsn) {
-		log_mutex_exit();
-		return(true);
-	}
-
-	if (log_sys.n_pending_checkpoint_writes > 0) {
-		/* A checkpoint write is running */
-		log_mutex_exit();
-
-		return(false);
-	}
-
-	log_sys.next_checkpoint_lsn = oldest_lsn;
-	log_write_checkpoint_info(end_lsn);
-	ut_ad(!log_mutex_own());
-
-	return(true);
-}
-
-/** Make a checkpoint */
-void log_make_checkpoint()
-{
-	/* Preflush pages synchronously */
-
-	while (!log_preflush_pool_modified_pages(LSN_MAX)) {
-		/* Flush as much as we can */
-	}
-
-	while (!log_checkpoint()) {
-		/* Force a checkpoint */
-	}
+	mysql_mutex_unlock(&log_sys.mutex);
 }
 
 /****************************************************************//**
-Tries to establish a big enough margin of free space in the log groups, such
+Tries to establish a big enough margin of free space in the log, such
 that a new log entry can be catenated without an immediate need for a
 checkpoint. NOTE: this function may only be called if the calling thread
 owns no synchronization objects! */
-static
-void
-log_checkpoint_margin(void)
-/*=======================*/
+ATTRIBUTE_COLD static void log_checkpoint_margin()
 {
-	ib_uint64_t	advance;
-	bool		success;
-loop:
-	advance = 0;
+  while (log_sys.check_flush_or_checkpoint())
+  {
+    mysql_mutex_lock(&log_sys.mutex);
+    ut_ad(!recv_no_log_write);
 
-	log_mutex_enter();
-	ut_ad(!recv_no_log_write);
+    if (!log_sys.check_flush_or_checkpoint())
+    {
+func_exit:
+      mysql_mutex_unlock(&log_sys.mutex);
+      return;
+    }
 
-	if (!log_sys.check_flush_or_checkpoint()) {
-		log_mutex_exit();
-		return;
-	}
+    const lsn_t lsn= log_sys.get_lsn();
+    const lsn_t checkpoint= log_sys.last_checkpoint_lsn;
+    const lsn_t sync_lsn= checkpoint + log_sys.max_checkpoint_age;
+    if (lsn <= sync_lsn)
+    {
+      log_sys.set_check_flush_or_checkpoint(false);
+      goto func_exit;
+    }
 
-	const lsn_t oldest_lsn = log_buf_pool_get_oldest_modification();
-	const lsn_t lsn = log_sys.get_lsn();
-	const lsn_t age = lsn - oldest_lsn;
+    mysql_mutex_unlock(&log_sys.mutex);
 
-	if (age > log_sys.max_modified_age_sync) {
-
-		/* A flush is urgent: we have to do a synchronous preflush */
-		advance = age - log_sys.max_modified_age_sync;
-	}
-
-	const lsn_t checkpoint_age = lsn - log_sys.last_checkpoint_lsn;
-
-	ut_ad(log_sys.max_checkpoint_age >= log_sys.max_checkpoint_age_async);
-	const bool do_checkpoint
-		= checkpoint_age > log_sys.max_checkpoint_age_async;
-
-	if (checkpoint_age <= log_sys.max_checkpoint_age) {
-		log_sys.set_check_flush_or_checkpoint(false);
-	}
-
-	log_mutex_exit();
-
-	if (advance) {
-		lsn_t	new_oldest = oldest_lsn + advance;
-
-		success = log_preflush_pool_modified_pages(new_oldest);
-
-		/* If the flush succeeded, this thread has done its part
-		and can proceed. If it did not succeed, there was another
-		thread doing a flush at the same time. */
-		if (!success) {
-			log_sys.set_check_flush_or_checkpoint();
-			goto loop;
-		}
-	}
-
-	if (do_checkpoint) {
-		log_checkpoint();
-	}
+    /* We must wait to prevent the tail of the log overwriting the head. */
+    buf_flush_wait_flushed(std::min(sync_lsn, checkpoint + (1U << 20)));
+    os_thread_sleep(10000); /* Sleep 10ms to avoid a thundering herd */
+  }
 }
 
 /**
@@ -1501,7 +978,7 @@ Checks that there is enough free space in the log to start a new query step.
 Flushes the log buffer or makes a new checkpoint if necessary. NOTE: this
 function may only be called if the calling thread owns no synchronization
 objects! */
-void log_check_margins()
+ATTRIBUTE_COLD void log_check_margins()
 {
   do
   {
@@ -1515,7 +992,7 @@ void log_check_margins()
 extern void buf_resize_shutdown();
 
 /** Make a checkpoint at the latest lsn on shutdown. */
-void logs_empty_and_mark_files_at_shutdown()
+ATTRIBUTE_COLD void logs_empty_and_mark_files_at_shutdown()
 {
 	lsn_t			lsn;
 	ulint			count = 0;
@@ -1541,7 +1018,6 @@ void logs_empty_and_mark_files_at_shutdown()
 		!srv_read_only_mode && srv_fast_shutdown < 2) {
 		buf_dump_start();
 	}
-	srv_error_monitor_timer.reset();
 	srv_monitor_timer.reset();
 	lock_sys.timeout_timer.reset();
 	if (do_srv_shutdown) {
@@ -1554,14 +1030,6 @@ loop:
 	ut_ad(log_sys.is_initialised() || !srv_was_started);
 	ut_ad(fil_system.is_initialised() || !srv_was_started);
 
-	if (!srv_read_only_mode) {
-		if (recv_sys.flush_start) {
-			/* This is in case recv_writer_thread was never
-			started, or buf_flush_page_cleaner
-			failed to notice its termination. */
-			os_event_set(recv_sys.flush_start);
-		}
-	}
 #define COUNT_INTERVAL 600U
 #define CHECK_INTERVAL 100000U
 	os_thread_sleep(CHECK_INTERVAL);
@@ -1623,50 +1091,14 @@ wait_suspend_loop:
 		goto wait_suspend_loop;
 	}
 
+	if (buf_page_cleaner_is_active) {
+		thread_name = "page cleaner thread";
+		pthread_cond_signal(&buf_pool.do_flush_list);
+		goto wait_suspend_loop;
+	}
+
 	buf_load_dump_end();
 
-	srv_shutdown_state = SRV_SHUTDOWN_FLUSH_PHASE;
-
-	/* At this point only page_cleaner should be active. We wait
-	here to let it complete the flushing of the buffer pools
-	before proceeding further. */
-
-	count = 0;
-	service_manager_extend_timeout(COUNT_INTERVAL * CHECK_INTERVAL/1000000 * 2,
-		"Waiting for page cleaner");
-	while (buf_page_cleaner_is_active) {
-		++count;
-		os_thread_sleep(CHECK_INTERVAL);
-		if (srv_print_verbose_log && count > COUNT_INTERVAL) {
-			service_manager_extend_timeout(COUNT_INTERVAL * CHECK_INTERVAL/1000000 * 2,
-				"Waiting for page cleaner");
-			ib::info() << "Waiting for page_cleaner to "
-				"finish flushing of buffer pool";
-			/* This is a workaround to avoid the InnoDB hang
-			when OS datetime changed backwards */
-			os_event_set(buf_flush_event);
-			count = 0;
-		}
-	}
-
-	if (log_sys.is_initialised()) {
-		log_mutex_enter();
-		const ulint	n_write	= log_sys.n_pending_checkpoint_writes;
-		const ulint	n_flush	= log_sys.pending_flushes;
-		log_mutex_exit();
-
-		if (n_write || n_flush) {
-			if (srv_print_verbose_log && count > 600) {
-				ib::info() << "Pending checkpoint_writes: "
-					<< n_write
-					<< ". Pending log flush writes: "
-					<< n_flush;
-				count = 0;
-			}
-			goto loop;
-		}
-	}
-
 	if (!buf_pool.is_initialised()) {
 		ut_ad(!srv_was_started);
 	} else if (ulint pending_io = buf_pool.io_pending()) {
@@ -1677,6 +1109,26 @@ wait_suspend_loop:
 		}
 
 		goto loop;
+	} else {
+		buf_flush_buffer_pool();
+	}
+
+	if (log_sys.is_initialised()) {
+		mysql_mutex_lock(&log_sys.mutex);
+		const ulint	n_write	= log_sys.n_pending_checkpoint_writes;
+		const ulint	n_flush	= log_sys.pending_flushes;
+		mysql_mutex_unlock(&log_sys.mutex);
+
+		if (n_write || n_flush) {
+			if (srv_print_verbose_log && count > 600) {
+				ib::info() << "Pending checkpoint_writes: "
+					<< n_write
+					<< ". Pending log flush writes: "
+					<< n_flush;
+				count = 0;
+			}
+			goto loop;
+		}
 	}
 
 	if (srv_fast_shutdown == 2 || !srv_was_started) {
@@ -1708,7 +1160,7 @@ wait_suspend_loop:
 			"ensuring dirty buffer pool are written to log");
 		log_make_checkpoint();
 
-		log_mutex_enter();
+		mysql_mutex_lock(&log_sys.mutex);
 
 		lsn = log_sys.get_lsn();
 
@@ -1717,7 +1169,7 @@ wait_suspend_loop:
 			+ SIZE_OF_FILE_CHECKPOINT;
 		ut_ad(lsn >= log_sys.last_checkpoint_lsn);
 
-		log_mutex_exit();
+		mysql_mutex_unlock(&log_sys.mutex);
 
 		if (lsn_changed) {
 			goto loop;
@@ -1774,17 +1226,22 @@ log_print(
 	double	time_elapsed;
 	time_t	current_time;
 
-	log_mutex_enter();
+	mysql_mutex_lock(&log_sys.mutex);
+
+	const lsn_t lsn= log_sys.get_lsn();
+	mysql_mutex_lock(&buf_pool.flush_list_mutex);
+	const lsn_t pages_flushed = buf_pool.get_oldest_modification(lsn);
+	mysql_mutex_unlock(&buf_pool.flush_list_mutex);
 
 	fprintf(file,
 		"Log sequence number " LSN_PF "\n"
 		"Log flushed up to   " LSN_PF "\n"
 		"Pages flushed up to " LSN_PF "\n"
 		"Last checkpoint at  " LSN_PF "\n",
-		log_sys.get_lsn(),
+		lsn,
 		log_sys.get_flushed_lsn(),
-		log_buf_pool_get_oldest_modification(),
-		log_sys.last_checkpoint_lsn);
+		pages_flushed,
+		lsn_t{log_sys.last_checkpoint_lsn});
 
 	current_time = time(NULL);
 
@@ -1809,7 +1266,7 @@ log_print(
 	log_sys.n_log_ios_old = log_sys.n_log_ios;
 	log_sys.last_printout_time = current_time;
 
-	log_mutex_exit();
+	mysql_mutex_unlock(&log_sys.mutex);
 }
 
 /**********************************************************************//**
@@ -1827,18 +1284,21 @@ void log_t::close()
 {
   ut_ad(this == &log_sys);
   if (!is_initialised()) return;
-  m_initialised = false;
+  m_initialised= false;
   log.close();
 
   ut_free_dodump(buf, srv_log_buffer_size);
-  buf = NULL;
+  buf= nullptr;
   ut_free_dodump(flush_buf, srv_log_buffer_size);
-  flush_buf = NULL;
+  flush_buf= nullptr;
 
-  mutex_free(&mutex);
-  mutex_free(&log_flush_order_mutex);
+  mysql_mutex_destroy(&mutex);
+  mysql_mutex_destroy(&flush_order_mutex);
 
   recv_sys.close();
+
+  aligned_free(checkpoint_buf);
+  checkpoint_buf= nullptr;
 }
 
 std::string get_log_file_path(const char *filename)
